# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: common/common.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x13\x63ommon/common.proto\x12\x06\x63ommon\"\xd3\x01\n\x04\x43ode\x12\x0e\n\x06s3_url\x18\x01 \x01(\t\x12\x0c\n\x04size\x18\x02 \x01(\x05\x12\x1f\n\x04type\x18\x03 \x01(\x0e\x32\x11.common.Code.Type\x12)\n\x07s3_meta\x18\x04 \x03(\x0b\x32\x18.common.Code.S3MetaEntry\x1a-\n\x0bS3MetaEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"2\n\x04Type\x12\x0b\n\x07NOT_SET\x10\x00\x12\x08\n\x04NBOX\x10\x01\x12\x07\n\x03ZIP\x10\x02\x12\n\n\x06\x42INARY\x10\x03\"\x8a\x01\n\x0bNBXAuthInfo\x12\x10\n\x08username\x18\x01 \x01(\t\x12\x14\n\x0cworkspace_id\x18\x02 \x01(\t\x12\x14\n\x0c\x61\x63\x63\x65ss_token\x18\x03 \x01(\t\x12\x15\n\rrefresh_token\x18\x04 \x01(\t\x12\x12\n\naccess_key\x18\x05 \x01(\t\x12\x12\n\nproject_id\x18\x06 \x01(\t\"S\n\x0fIntegrationData\x12\x15\n\rconnection_id\x18\x01 \x01(\t\x12\x17\n\x0f\x63onnection_name\x18\x02 \x01(\t\x12\x10\n\x08infra_id\x18\x03 \x01(\t\"\x80\x01\n\x08Resource\x12\x0b\n\x03\x63pu\x18\x01 \x01(\t\x12\x0e\n\x06memory\x18\x02 \x01(\t\x12\x0b\n\x03gpu\x18\x03 \x01(\t\x12\x11\n\tgpu_count\x18\x04 \x01(\t\x12\x11\n\tdisk_size\x18\x05 \x01(\t\x12\x0f\n\x07timeout\x18\x06 \x01(\x03\x12\x13\n\x0bmax_retries\x18\x07 \x01(\x05\"N\n\x03Log\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0b\n\x03idx\x18\x02 \x01(\x05\x12\x0b\n\x03log\x18\x03 \x01(\t\x12!\n\x08k8s_info\x18\x04 \x01(\x0b\x32\x0f.common.K8sInfo\"[\n\x07K8sInfo\x12\x0c\n\x04pods\x18\x01 \x03(\t\x12\x12\n\ndeployment\x18\x02 \x01(\t\x12\x0f\n\x07service\x18\x03 \x01(\t\x12\x0b\n\x03hpa\x18\x04 \x01(\t\x12\x10\n\x08\x63ron_job\x18\x05 \x01(\tBAZ?github.com/NimbleBoxAI/jobs-architecture/common_pb/v3;common_pbb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'common.common_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z?github.com/NimbleBoxAI/jobs-architecture/common_pb/v3;common_pb'
  _CODE_S3METAENTRY._options = None
  _CODE_S3METAENTRY._serialized_options = b'8\001'
  _CODE._serialized_start=32
  _CODE._serialized_end=243
  _CODE_S3METAENTRY._serialized_start=146
  _CODE_S3METAENTRY._serialized_end=191
  _CODE_TYPE._serialized_start=193
  _CODE_TYPE._serialized_end=243
  _NBXAUTHINFO._serialized_start=246
  _NBXAUTHINFO._serialized_end=384
  _INTEGRATIONDATA._serialized_start=386
  _INTEGRATIONDATA._serialized_end=469
  _RESOURCE._serialized_start=472
  _RESOURCE._serialized_end=600
  _LOG._serialized_start=602
  _LOG._serialized_end=680
  _K8SINFO._serialized_start=682
  _K8SINFO._serialized_end=773
# @@protoc_insertion_point(module_scope)
