from nbox.plugins.base import experimental_warning

experimental_warning("tensorboard")

def load_tensorboard_to_exp():
  """Experimental feature to load your tensorboard logs to nbox LMAO"""
  raise NotImplementedError("This feature is not yet implemented")
