# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: rsock.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='rsock.proto',
  package='rsock',
  syntax='proto3',
  serialized_options=b'ZBgithub.com/NimbleBoxAI/nimblebox-rsock-server/v1/rsock_pb;rsock_pb',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x0brsock.proto\x12\x05rsock\"U\n\x10HandshakeRequest\x12\x0b\n\x03jwt\x18\x01 \x01(\t\x12\x13\n\x0binstanceKey\x18\x02 \x01(\t\x12\x11\n\tlaunchURL\x18\x03 \x01(\t\x12\x0c\n\x04port\x18\x04 \x01(\t\"\"\n\x11HandshakeResponse\x12\r\n\x05token\x18\x01 \x01(\t\")\n\nDataPacket\x12\r\n\x05token\x18\x01 \x01(\t\x12\x0c\n\x04\x64\x61ta\x18\x02 \x01(\x0c\x32{\n\x05RSock\x12>\n\tHandshake\x12\x17.rsock.HandshakeRequest\x1a\x18.rsock.HandshakeResponse\x12\x32\n\x06Tunnel\x12\x11.rsock.DataPacket\x1a\x11.rsock.DataPacket(\x01\x30\x01\x42\x44ZBgithub.com/NimbleBoxAI/nimblebox-rsock-server/v1/rsock_pb;rsock_pbb\x06proto3'
)




_HANDSHAKEREQUEST = _descriptor.Descriptor(
  name='HandshakeRequest',
  full_name='rsock.HandshakeRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='jwt', full_name='rsock.HandshakeRequest.jwt', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='instanceKey', full_name='rsock.HandshakeRequest.instanceKey', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='launchURL', full_name='rsock.HandshakeRequest.launchURL', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='port', full_name='rsock.HandshakeRequest.port', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=22,
  serialized_end=107,
)


_HANDSHAKERESPONSE = _descriptor.Descriptor(
  name='HandshakeResponse',
  full_name='rsock.HandshakeResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='token', full_name='rsock.HandshakeResponse.token', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=109,
  serialized_end=143,
)


_DATAPACKET = _descriptor.Descriptor(
  name='DataPacket',
  full_name='rsock.DataPacket',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='token', full_name='rsock.DataPacket.token', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='data', full_name='rsock.DataPacket.data', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=145,
  serialized_end=186,
)

DESCRIPTOR.message_types_by_name['HandshakeRequest'] = _HANDSHAKEREQUEST
DESCRIPTOR.message_types_by_name['HandshakeResponse'] = _HANDSHAKERESPONSE
DESCRIPTOR.message_types_by_name['DataPacket'] = _DATAPACKET
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

HandshakeRequest = _reflection.GeneratedProtocolMessageType('HandshakeRequest', (_message.Message,), {
  'DESCRIPTOR' : _HANDSHAKEREQUEST,
  '__module__' : 'rsock_pb2'
  # @@protoc_insertion_point(class_scope:rsock.HandshakeRequest)
  })
_sym_db.RegisterMessage(HandshakeRequest)

HandshakeResponse = _reflection.GeneratedProtocolMessageType('HandshakeResponse', (_message.Message,), {
  'DESCRIPTOR' : _HANDSHAKERESPONSE,
  '__module__' : 'rsock_pb2'
  # @@protoc_insertion_point(class_scope:rsock.HandshakeResponse)
  })
_sym_db.RegisterMessage(HandshakeResponse)

DataPacket = _reflection.GeneratedProtocolMessageType('DataPacket', (_message.Message,), {
  'DESCRIPTOR' : _DATAPACKET,
  '__module__' : 'rsock_pb2'
  # @@protoc_insertion_point(class_scope:rsock.DataPacket)
  })
_sym_db.RegisterMessage(DataPacket)


DESCRIPTOR._options = None

_RSOCK = _descriptor.ServiceDescriptor(
  name='RSock',
  full_name='rsock.RSock',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=188,
  serialized_end=311,
  methods=[
  _descriptor.MethodDescriptor(
    name='Handshake',
    full_name='rsock.RSock.Handshake',
    index=0,
    containing_service=None,
    input_type=_HANDSHAKEREQUEST,
    output_type=_HANDSHAKERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Tunnel',
    full_name='rsock.RSock.Tunnel',
    index=1,
    containing_service=None,
    input_type=_DATAPACKET,
    output_type=_DATAPACKET,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_RSOCK)

DESCRIPTOR.services_by_name['RSock'] = _RSOCK

# @@protoc_insertion_point(module_scope)
